"""Initial waitlist construction."""

import math
import ciw


def begin_service_if_possible_accept(node, next_individual):
    """
    Begins the service of the next individual (at acceptance point):
        - Sets the arrival date as the current time
        - If there is a free server or there are infinite servers:
            - Attach the server to the individual (only when servers are not infinite)
        - Get service start time, service time, service end time
        - Update the server's end date (only when servers are not infinite)
    """
    free_server = node.find_free_server(next_individual)
    if free_server is None and math.isinf(node.c) is False:
        node.decide_preempt(next_individual)
    if free_server is not None or math.isinf(node.c):
        if math.isinf(node.c) is False:
            node.attach_server(free_server, next_individual)
        next_individual.service_start_date = 0
        next_individual.service_time = node.get_service_time(next_individual)
        next_individual.service_end_date = next_individual.service_time
        if not math.isinf(node.c):
            free_server.next_end_service_date = next_individual.service_end_date


# from the ciw source code; needs to be adapted to accommodate the pre-existing wait list
def accept(individual_id, individual_class, arrival_date, node, simulation):
    """
    Accepts a new customer to the queue:
      - record all other information at arrival point
      - update state tracker
    """
    simulation.current_time = arrival_date
    next_individual = ciw.Individual(
        id_number=individual_id,
        customer_class=individual_class,
        priority_class=simulation.network.priority_class_mapping[individual_class],
        simulation=simulation,
    )
    next_individual.node = node.id_number
    next_individual.queue_size_at_arrival = "Unknown"
    node.individuals[next_individual.priority_class].append(next_individual)
    node.number_of_individuals += 1
    node.simulation.statetracker.change_state_accept(node, next_individual)
    next_individual.arrival_date = arrival_date
    begin_service_if_possible_accept(node, next_individual)

    simulation.nodes[0].number_of_individuals += 1
    simulation.nodes[0].number_of_individuals_per_class[
        next_individual.customer_class
    ] += 1
    simulation.nodes[0].number_accepted_individuals += 1
    simulation.nodes[0].number_accepted_individuals_per_class[
        next_individual.customer_class
    ] += 1


def create_existing_customers_from_list(backlog, Q):
    """Occupies instance of simulation with individuals.


    PARAMETERS
    ----------
    backlog (list): List of data for each individual to be added.
    Q (ciw.Simulation): Instance of simulation.
    """
    customer_count = 1
    for row in backlog:
        customer_id = row[0]
        customer_class = row[1]
        customer_arrival_date = row[2]
        customer_node = row[3]
        accept(
            customer_id,
            customer_class,
            customer_arrival_date,
            Q.nodes[customer_node],
            Q,
        )
        customer_count += 1



import ciw

# https://en.wikipedia.org/wiki/Adversarial_queueing_network
# https://web.mst.edu/~gosavia/queuing_formulas.pdf
# https://en.wikipedia.org/wiki/Retrial_queue
# https://en.wikipedia.org/wiki/Little%27s_law
# https://en.wikipedia.org/wiki/Queuing_Rule_of_Thumb
# https://en.wikipedia.org/wiki/Queueing_theory


class GordonNewellNetwork:
    """https://en.wikipedia.org/wiki/Gordon%E2%80%93Newell_theorem"""

    def __init__(self):
        raise NotImplementedError


class JacksonNetwork:
    """https://en.wikipedia.org/wiki/Jackson_network"""

    def __init__(self):
        raise NotImplementedError


class MMcQueue:
    """https://en.wikipedia.org/wiki/M/M/c_queue"""

    def __init__(self):
        raise NotImplementedError


class BCMPNetwork:
    """https://en.wikipedia.org/wiki/BCMP_network"""

    def __init__(self):
        raise NotImplementedError


class GelenbeNetwork:
    """https://en.wikipedia.org/wiki/G-network"""

    def __init__(self):
        raise NotImplementedError


class KellyNetwork:
    """https://en.wikipedia.org/wiki/Kelly_network"""

    def __init__(self):
        raise NotImplementedError


class DM1Queue:
    """https://en.wikipedia.org/wiki/D/M/1_queue"""

    def __init__(self):
        raise NotImplementedError


class MD1Queue:
    """https://en.wikipedia.org/wiki/M/D/1_queue"""

    def __init__(self):
        raise NotImplementedError


class MDCQueue:
    """https://en.wikipedia.org/wiki/M/D/c_queue"""

    def __init__(self):
        raise NotImplementedError


class MM1Queue:
    """https://en.wikipedia.org/wiki/M/M/1_queue"""

    def __init__(self):
        raise NotImplementedError


class MG1Queue:
    """https://en.wikipedia.org/wiki/M/G/1_queue"""

    def __init__(self):
        raise NotImplementedError


class GM1Queue:
    """https://en.wikipedia.org/wiki/G/M/1_queue"""

    def __init__(self):
        raise NotImplementedError


class MGkQueue:
    """https://en.wikipedia.org/wiki/M/G/k_queue"""

    def __init__(self):
        raise NotImplementedError


class GG1Queue:
    """https://en.wikipedia.org/wiki/G/G/1_queue"""

    def __init__(self):
        raise NotImplementedError


class ForkJoinQueue:
    """https://en.wikipedia.org/wiki/Fork%E2%80%93join_queue"""

    def __init__(self):
        raise NotImplementedError


class BulkQueue:
    """https://en.wikipedia.org/wiki/Bulk_queue"""

    def __init__(self):
        raise NotImplementedError



"""Server schedules or factories of schedules."""

# TODO: Remove holidays as days where service is given.
def server_schedule_from_date_range(start_date, end_date):
    dates = pd.date_range(start_date, end_date)
    idx = list(range(1, dates.size + 1))
    is_weekday = (dates.weekday <= 4) * 1
    # Check which dates are weekdays
    return list(zip(is_weekday, idx))



"""Custom distributions to be used with Ciw."""
import ciw

class SeqPMFNaive(ciw.dists.Distribution):
    """Samples from a sequence of PMFs, else a naive extrapolation assumption.

    Input times are rounded.

    The naive forecast assumption is that the probability mass function in the future of the sequence is the same as the last element of the sequence.
    The naive backcast assumption is that the probability mass function in the past of the sequence is the same as the first element of the sequence.
    These naive extrapolation assumptions are not intended
    """

    def __init__(self, distseq):
        if not all(map(lambda o: isinstance(o, ciw.dists.Pmf)), distseq):
            raise ValueError("Not all inputs of distseq were of type ciw.dists.Pmf")
        self.distseq = distdeq

    def sample(self, t, ind=None):
        time = round(t)  # TODO: Remove once probability interpolation is implemented.
        if time in self.distseq:
            return self.distseq[time].sample(time, ind)
        elif time > max(self.distseq):
            return self.distseq[max(self.distseq)].sample(time, ind)
        elif time < min(self.distseq):
            return self.distseq[min(self.distseq)].sample(time, ind)
        else:  # TODO: Linearly interpolate probabilities wrt t
            raise ValueError(f"Unsupported sampling time of {t} in SeqPMFNaive.")


class DeterministicSeqNaive(ciw.dists.Distribution):
    """Samples from a sequence of constant random variables."""

    def __init__(self, distseq):
        if not all(map(lambda o: isinstance(o, ciw.dists.Deterministic)), distseq):
            raise ValueError(
                "Not all inputs of distseq were of type ciw.dists.Deterministic"
            )
        self.distseq = distseq

    def sample(self, t, ind=None):
        time = round(t)
        if time in self.disteq:
            return self.distseq[time].sample(time, ind)
        elif time > max(self.distseq):
            return self.distseq[max(self.distseq)].sample(time, ind)
        elif time < min(self.distseq):
            return self.distseq[min(self.distseq)].sample(time, ind)
        else:
            raise ValueError(
                f"Unsupported sampling time of {t} in DeterministicSeqNaive."
            )


def arrival_dists_from_grouped_prophet_forecast(forecasts, group_column="GROUPS"):
    """Takes the concatenated predictions"""
    dists = {}

    dists["yhat"] = {}
    dists["yhat_upper"] = {}
    dists["yhat_lower"] = {}

    for facility, facility_demand_df in forecasts.grouby(by=group_column):
        dists["yhat"][facility] = ciw.dists.Sequential(
            sequence=facility_demand_df["yhat"].to_list()
        )
        dists["yhat_upper"][facility] = ciw.dists.Sequential(
            sequence=facility_demand_df["yhat_upper"].to_list()
        )
        dists["yhat_lower"][facility] = ciw.dists.Sequential(
            sequence=facility_demand_df["yhat_lower"].to_list()
        )

    return dists
